// 002 u_c_2d_c_point_list
// 14 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

// -- uses a tStringList

(*$r+*)

unit u_c_2d_c_point_list;
  interface
    uses Classes, Graphics, u_c_basic_object
        ;

    type c_2d_point= Class(c_basic_object)
                       m_x, m_y: Double;

                       constructor create_2d_point(p_name: String; p_x, p_y: Double);
                       function f_display_2d_point: String; Virtual;
                       procedure display_2d_point; Virtual;
                       function f_c_self: c_2d_point;
                       procedure draw_2d_point(p_c_canvas: tCanvas; p_pen_color, p_radius: Integer); Virtual;
                     end; // c_2d_point

         c_2d_point_list= // "2d_point" list
                          Class(c_basic_object)
                            _m_c_2d_point_list: tStringList;
                            m_is_closed: Boolean;

                            Constructor create_2d_point_list(p_name: String);

                            function f_2d_point_count: Integer;
                            function f_c_2d_point(p_2d_point_index: Integer): c_2d_point;
                            function f_index_of(p_2d_point_name: String): Integer;
                            function f_index_of_c_2d_point(p_c_2d_point: c_2d_point): Integer;
                            procedure add_2d_point(p_2d_point_name: String; p_c_2d_point: c_2d_point);
                            procedure insert_2d_point(p_2d_point_name: String; p_c_2d_point: c_2d_point;
                                p_insertion_point: Integer);
                            procedure delete_2d_c_point(p_c_2d_point: c_2d_point);
                            procedure add_and_sort_x_2d_point(p_2d_point_name: String; p_c_2d_point: c_2d_point);

                            procedure display_2d_point_list; Virtual;

                            procedure draw_2d_point_list_poly_line(p_c_canvas: tCanvas;
                                p_pen_color: Integer);
                            procedure draw_2d_point_list_points(p_c_canvas: tCanvas;
                                p_pen_color, p_radius: Integer);

                            Destructor Destroy; Override;
                          end; // c_2d_point_list

  implementation
    uses SysUtils, u_c_display, u_double;

    // -- c_2d_point

    constructor c_2d_point.create_2d_point(p_name: String; p_x, p_y: Double);
      begin
        Inherited create_basic_object(p_name);
        m_x:= p_x; m_y:= p_y;
      end; // create_2d_point

    function c_2d_point.f_display_2d_point: String;
      begin
        Result:= m_name+ ' '+ f_display_double(m_x, 3, 2)+ ' '+ f_display_double(m_y, 3, 2);
      end; // f_display_2d_point

    procedure c_2d_point.display_2d_point;
      begin
        display(f_display_2d_point);
      end; // display_2d_point

    function c_2d_point.f_c_self: c_2d_point;
      begin
        Result:= Self;
      end; // f_c_self

    procedure c_2d_point.draw_2d_point(p_c_canvas: tCanvas; p_pen_color, p_radius: Integer);
      var l_save_pen_color: Integer;
      begin
        with p_c_canvas do
        begin
          l_save_pen_color:= Pen.Color;
          Pen.Color:= p_pen_color;

          Ellipse(Round(m_x)- p_radius, Round(m_y)- p_radius, Round(m_x)+ p_radius, Round(m_y)+ p_radius);

          Pen.Color:= l_save_pen_color;
        end; // with p_c_canvas
      end; // draw_2d_point

    // -- c_2d_point_list

    Constructor c_2d_point_list.create_2d_point_list(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        _m_c_2d_point_list:= tStringList.Create;
      end; // create_2d_point_line

    function c_2d_point_list.f_2d_point_count: Integer;
      begin
        Result:= _m_c_2d_point_list.Count;
      end; // f_2d_point_count

    function c_2d_point_list.f_c_2d_point(p_2d_point_index: Integer): c_2d_point;
      begin
        // -- WRAP AROUND: for Count, return point 0
        if (p_2d_point_index= f_2d_point_count) and m_is_closed
          then Result:= c_2d_point(_m_c_2d_point_list.Objects[0])
          else Result:= c_2d_point(_m_c_2d_point_list.Objects[p_2d_point_index]);
      end; //  f_2d_point

    function c_2d_point_list.f_index_of(p_2d_point_name: String): Integer;
      begin
        Result:= _m_c_2d_point_list.IndexOf(p_2d_point_name);
      end; // f_index_of

    function c_2d_point_list.f_index_of_c_2d_point(p_c_2d_point: c_2d_point): Integer;
      var l_index_of: Integer;
      begin
        l_index_of:= -1;
        for l_index_of:= 0 to f_2d_point_count- 1 do
          if f_c_2d_point(l_index_of)= p_c_2d_point
            then begin
                Result:= l_index_of;
              end;
      end; // f_index_of_c_2d_point

    procedure c_2d_point_list.add_2d_point(p_2d_point_name: String; p_c_2d_point: c_2d_point);
      begin
        _m_c_2d_point_list.AddObject(p_2d_point_name, p_c_2d_point);
      end; // add_2d_point

    procedure c_2d_point_list.insert_2d_point(p_2d_point_name: String; p_c_2d_point: c_2d_point;
        p_insertion_point: Integer);
      begin
        _m_c_2d_point_list.InsertObject(p_insertion_point, p_2d_point_name, p_c_2d_point);
      end; // insert_2d_point

    procedure c_2d_point_list.delete_2d_c_point(p_c_2d_point: c_2d_point);
      var l_index_of: Integer;
      begin
        // -- find the point
        l_index_of:= f_index_of_c_2d_point(p_c_2d_point);
        if l_index_of>= 0
          then begin
              f_c_2d_point(l_index_of).Free;
              _m_c_2d_point_list.delete(l_index_of);
            end;
      end; // delete_2d_c_point

    procedure c_2d_point_list.add_and_sort_x_2d_point(p_2d_point_name: String; p_c_2d_point: c_2d_point);
      var l_insertion_index: Integer;
      begin
        // -- find the insertion point
        l_insertion_index:= 0;
        while l_insertion_index< f_2d_point_count do
          if f_c_2d_point(l_insertion_index).m_x> p_c_2d_point.m_x
            then break
            else Inc(l_insertion_index);

        if l_insertion_index= f_2d_point_count
          then begin
              add_2d_point(p_2d_point_name, p_c_2d_point)
            end
          else begin
              insert_2d_point(p_2d_point_name, p_c_2d_point, l_insertion_index);
            end;
      end; // add_and_sort_x_2d_point

    procedure c_2d_point_list.display_2d_point_list;
      var l_2d_point_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_2d_point_count));

        for l_2d_point_index:= 0 to f_2d_point_count- 1 do
          f_c_2d_point(l_2d_point_index).display_2d_point;
      end; // display_2d_point_list

    procedure c_2d_point_list.draw_2d_point_list_poly_line(p_c_canvas: tCanvas;
        p_pen_color: Integer);
      var l_save_pen_color: Integer;
          l_2d_point_index: Integer;
      begin
        if f_2d_point_count> 0
          then
            with p_c_canvas do
            begin
              l_save_pen_color:= Pen.Color;
              Pen.Color:= p_pen_color;

              with f_c_2d_point(0) do
                MoveTo(Round(m_x), Round(m_y));
              for l_2d_point_index:= 1 to f_2d_point_count- 1 do
                with f_c_2d_point(l_2d_point_index) do
                  LineTo(Round(m_x), Round(m_y));

              Pen.Color:= l_save_pen_color;
            end; // with p_c_canvas
      end; // draw_2d_point_list_poly_line

    procedure c_2d_point_list.draw_2d_point_list_points(p_c_canvas: tCanvas;
        p_pen_color, p_radius: Integer);
      var l_2d_point_index: Integer;
      begin
        with p_c_canvas do
          for l_2d_point_index:= 0 to f_2d_point_count- 1 do
            f_c_2d_point(l_2d_point_index).draw_2d_point(p_c_canvas, p_pen_color, p_radius);
      end; // draw_2d_point_list_points

    Destructor c_2d_point_list.Destroy;
      var l_2d_point_index: Integer;
      begin
        for l_2d_point_index:= 0 to f_2d_point_count- 1 do
          f_c_2d_point(l_2d_point_index).Free;
        _m_c_2d_point_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_2d_point_list
    end. // u_c_2d_point_list
