// 002 u_c_2d_selection_point_list
// 14 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

// -- add a m_selected field to the points and
// --   selection functions to the list

(*$r+*)

unit u_c_2d_selection_point_list;
  interface
    uses Classes, Graphics, u_c_basic_object
        , u_c_2d_c_point_list
        ;

    const k_selection_distance_max= 5;

    type c_2d_selection_point= Class(c_2d_point)
                                 m_is_selected: Boolean;
                                 m_radius: Integer;

                                 constructor create_2d_selection_point(p_name: String; p_x, p_y: Double;
                                     p_radius: Integer);
                                 function f_c_self: c_2d_selection_point;
                                 function f_display_2d_point: String; Override;
                                 procedure display_2d_point; Override;
                                 procedure draw_2d_selection_point(p_c_canvas: tCanvas;
                                     p_pen_color: Integer);
                               end; // c_2d_selection_point

         c_2d_selection_point_list= // "2d_point" list
                                    Class(c_2d_point_list)
                                      m_c_2d_selected_point_ref: c_2d_selection_point;
                                      m_c_canvas_ref: tCanvas;

                                      Constructor create_2d_selection_point_list(p_name: String;
                                         p_c_canvas_ref: tCanvas);
                                      procedure draw_selected_point;
                                      function f_c_nearest_point(p_x, p_y: Double): c_2d_selection_point;
                                      procedure display_2d_point_list; Override;
                                      Destructor Destroy; Override;
                                    end; // c_2d_selection_point_list

  implementation
    uses SysUtils, u_c_display, u_double;

    // -- c_2d_selection_point

    constructor c_2d_selection_point.create_2d_selection_point(p_name: String; p_x, p_y: Double;
        p_radius: Integer);
      begin
        Inherited create_basic_object(p_name);
        m_x:= p_x; m_y:= p_y;
        m_radius:= p_radius;
      end; // create_2d_selection_point

    function c_2d_selection_point.f_c_self: c_2d_selection_point;
      begin
        Result:= Self;
      end; // f_c_self

    function c_2d_selection_point.f_display_2d_point: String;
      begin
        if m_is_selected
          then Result:= m_name+ ' '+ f_display_double(m_x, 3, 2)+ ' '+ f_display_double(m_y, 3, 2)+ ' SEL'
          else Result:= Inherited f_display_2d_point;
      end; // f_display_2d_point

    procedure c_2d_selection_point.display_2d_point;
      begin
        display(f_display_2d_point);
      end; // display_2d_point

    procedure c_2d_selection_point.draw_2d_selection_point(p_c_canvas: tCanvas;
        p_pen_color: Integer);
      begin
        if m_is_selected and (p_pen_color<> clWhite)
          then draw_2d_point(p_c_canvas, clRed, m_radius)
          else draw_2d_point(p_c_canvas, p_pen_color, m_radius)
      end; // draw_2d_point

    // -- c_2d_selection_point_list

    Constructor c_2d_selection_point_list.create_2d_selection_point_list(p_name: String;
        p_c_canvas_ref: tCanvas);
      begin
        Inherited create_2d_point_list(p_name);
        m_c_canvas_ref:= p_c_canvas_ref;
      end; // create_2d_selection_point_list

    procedure c_2d_selection_point_list.draw_selected_point;
      begin
      end; // draw_selected_point

    function c_2d_selection_point_list.f_c_nearest_point(p_x, p_y: Double): c_2d_selection_point;
      var l_distance_min, l_distance: Double;
          l_2d_point_index: Integer;
          l_c_2d_nearest_point: c_2d_point;
      begin
        Result:= Nil;

        if f_2d_point_count> 0
          then begin
              with f_c_2d_point(0) do
                l_distance_min:= Sqrt(Sqr(p_x- m_x)+ Sqr(p_y- m_y));

              l_c_2d_nearest_point:= f_c_2d_point(0);

              for l_2d_point_index:= 1 to f_2d_point_count- 1 do
                with f_c_2d_point(l_2d_point_index) do
                begin
                  l_distance:= Sqrt(Sqr(p_x- m_x)+ Sqr(p_y- m_y));
                  if l_distance< l_distance_min
                    then begin
                        l_distance_min:= l_distance;
                        l_c_2d_nearest_point:= f_c_self;
                      end;
                end;

              if (Abs(l_c_2d_nearest_point.m_x- p_x)+ Abs(l_c_2d_nearest_point.m_y- p_y)< k_selection_distance_max)
                then Result:= c_2d_selection_point(l_c_2d_nearest_point);
            end;

        if Result= Nil
          then begin
              if m_c_2d_selected_point_ref<> Nil
                then begin
                    m_c_2d_selected_point_ref.m_is_selected:= False;
                    m_c_2d_selected_point_ref.draw_2d_selection_point(m_c_canvas_ref, clBlack);
                    // -- nil the Reference
                    m_c_2d_selected_point_ref:= Nil;
                  end;
            end
          else begin
              if m_c_2d_selected_point_ref<> Nil
                then begin
                    m_c_2d_selected_point_ref.m_is_selected:= False;
                    m_c_2d_selected_point_ref.draw_2d_selection_point(m_c_canvas_ref, clBlack);
                  end;

               m_c_2d_selected_point_ref:= Result;
               m_c_2d_selected_point_ref.m_is_selected:= True;

               m_c_2d_selected_point_ref.draw_2d_selection_point(m_c_canvas_ref, clRed);
            end;
      end; // f_nearest_point_index

    procedure c_2d_selection_point_list.display_2d_point_list;
      begin
        Inherited;
        if m_c_2d_selected_point_ref<> Nil
          then m_c_2d_selected_point_ref.display_2d_point;
      end; // display_2d_point_l ist

    Destructor c_2d_selection_point_list.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_2d_selection_point_list
    end. // u_c_2d_selection_point_list
