// 002 u_c_rectangle_selecter
// 14 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_rectangle_selecter;
  interface
    uses Graphics
        , Classes // tShiftState
        , Controls // tMouseButton
        , StdCtrls
        , ComCtrls
        , ExtCtrls // tPanel, tPaintBox
        , u_c_basic_object;

    type t_selection_type= (e_unknown_selection,
             e_north_west_selection, e_north_east_selection,
             e_south_east_selection, e_south_west_selection,
             e_north_selection, e_east_selection, e_south_selection, e_west_selection,
             e_inside_selection);

         c_rectangle_selecter= class(c_basic_object)
                                 m_c_paintbox_ref: tPaintBox;

                                 m_is_active: Boolean;
                                 m_selection_type: t_selection_type;
                                 m_shrink_if_outside_click: Boolean;
                                 m_selection_distance_max: Integer;

                                 m_x_1, m_y_1, m_x_2, m_y_2: Integer;
                                 m_previous_x_1, m_previous_y_1, m_previous_x_2, m_previous_y_2: Integer;

                                 // -- for moving (Right click)
                                 m_mouse_x, m_mouse_y: Integer;

                                 Constructor create_rectangle_selecter(p_name: String;
                                     p_c_paintbox_ref: tPaintbox;
                                     p_selection_distance_max: Integer);

                                 function f_is_empty: Boolean;
                                 procedure empty_selecter;
                                 procedure draw_selecter;

                                 function f_handled_mouse_down(
                                     p_mouse_button: TMouseButton;
                                     p_shift_state: TShiftState; p_x, p_y: Integer): Boolean;
                                 function f_handled_mouse_move(
                                     p_shift_state: TShiftState; p_x, p_y: Integer): Boolean;
                                 function f_handled_mouse_up(
                                     p_mouse_button: TMouseButton;
                                     p_shift_state: TShiftState; p_x, p_y: Integer): Boolean;

                                 function f_is_in_selecter(p_x, p_y: Integer): Boolean;
                                 procedure move_point(var pv_x, pv_y: Double);
                               end; // c_rectangle_selecter

  implementation
    Uses SysUtils, TypInfo
         , u_c_display, u_strings, u_double;

    function f_selection_type_name(p_selection_type: t_selection_type): String;
      begin
        Result:= GetEnumName(TypeInfo(t_selection_type), Integer(p_selection_type));
        Delete(Result, 1, 2);
        Delete(Result, Length(Result)- Length('_selection')+ 1, Length('_selection'));
      end; // f_selection_type_name

    // -- c_rectangle_selecter

    Constructor c_rectangle_selecter.create_rectangle_selecter(p_name: String;
        p_c_paintbox_ref: tPaintbox;
        p_selection_distance_max: Integer);
      begin
        Inherited create_basic_object(p_name);

        m_c_paintbox_ref:= p_c_paintbox_ref;
        m_selection_distance_max:= p_selection_distance_max;
      end; // create_rectangle_selecter

    function c_rectangle_selecter.f_is_empty: Boolean;
      begin
        Result:= Abs(m_x_1)+ Abs(m_y_1)+ Abs(m_x_2)+ Abs(m_y_2)= 0
      end; // f_is_empty

    procedure c_rectangle_selecter.empty_selecter;
      begin
        m_x_1:= 0; m_y_1:= 0; m_x_2:= 0; m_y_2:= 0;
      end; // empty_selecter

    procedure c_rectangle_selecter.draw_selecter;
      var l_save_brush_style: tBrushStyle;
          l_save_pen_style: tPenStyle;
          l_save_pen_mode: tPenMode;
      begin
        if f_is_empty
          then Exit;

        with m_c_paintbox_ref.Canvas do
        begin
          l_save_brush_style:= Brush.Style;
          l_save_pen_style:= Pen.Style;
          l_save_pen_mode:= Pen.Mode;

          Brush.Style:= bsClear;
          // -- cannot color
          // Pen.Color:= clRed;
          Pen.Style:= psDot;
          Pen.Mode:= pmNot;

          Rectangle(m_x_1, m_y_1, m_x_2, m_y_2);

          Brush.Style:= l_save_brush_style;
          Pen.Style:= l_save_pen_style;
          Pen.Mode:= l_save_pen_mode;
        end; // with m_c_paintbox_ref.Canvas
      end; // draw_selecter

    function c_rectangle_selecter.f_handled_mouse_down(p_mouse_button: TMouseButton;
        p_shift_state: TShiftState; p_x, p_y: Integer): Boolean;
      var l_distance_min: Integer;

      procedure check_selection(p_corner_x, p_corner_y: Integer; p_selection_type: t_selection_type);
        var l_corner_distance: Integer;
        begin
          l_corner_distance:= Abs(p_x- p_corner_x)+ Abs(p_y- p_corner_y);
          if l_corner_distance< l_distance_min
            then begin
                m_selection_type:= p_selection_type;
                l_distance_min:= l_corner_distance;
              end;
        end; // check_selection

      function f_check_horizontal_border(p_border_y: Integer; p_selection_type: t_selection_type): Boolean;
        begin
          Result:= (p_x>= m_x_1+ m_selection_distance_max) and (p_x< m_x_2- m_selection_distance_max)
              and (p_y>= p_border_y- m_selection_distance_max) and (p_y<= p_border_y+ m_selection_distance_max);
          if Result
            then m_selection_type:= p_selection_type;
        end; // f_check_horizontal_border

      function f_check_vertical_border(p_border_x: Integer; p_selection_type: t_selection_type): Boolean;
        begin
          Result:= (p_y>= m_y_1+ m_selection_distance_max) and (p_y< m_y_2- m_selection_distance_max)
              and (p_x>= p_border_x- m_selection_distance_max) and (p_x<= p_border_x+ m_selection_distance_max);
          if Result
            then m_selection_type:= p_selection_type;
        end; // f_check_horizontal_border

      begin // f_handled_mouse_down
        Result:= m_is_active;

        if Result
          then begin
              m_mouse_x:= p_x; m_mouse_y:= p_y;

              // -- when all nil, allow south_west drag
              if f_is_empty
                then m_selection_type:= e_south_east_selection
                else begin
                    // -- compute distance to 4 corners
                    // --  (this computes the nearest corner. Could be
                    // --   more restrictive and only test vicinity)
                    l_distance_min:= Abs(p_x- m_x_1)+ Abs(p_y- m_y_1);
                    m_selection_type:= e_north_west_selection;

                    check_selection(m_x_2, m_y_1, e_north_east_selection);
                    check_selection(m_x_2, m_y_2, e_south_east_selection);
                    check_selection(m_x_1, m_y_2, e_south_west_selection);

                    if l_distance_min< m_selection_distance_max
                      then
                      else begin
                          // -- check borders
                          if    f_check_horizontal_border(m_y_1, e_north_selection)
                              or
                                f_check_horizontal_border(m_y_2, e_south_selection)
                              or
                                f_check_vertical_border(m_x_1, e_west_selection)
                              or
                                 f_check_vertical_border(m_x_2, e_east_selection)
                            then 
                            else
                              if (p_x>= m_x_1+ m_selection_distance_max) and (p_x< m_x_2- m_selection_distance_max)
                                  and (p_y>= m_y_1+ m_selection_distance_max) and (p_y< m_y_2- m_selection_distance_max)
                                then m_selection_type:= e_inside_selection
                                else m_selection_type:= e_unknown_selection
                        end;

                    display(f_selection_type_name(m_selection_type)+ ' '+ IntToStr(l_distance_min));

                    if m_shrink_if_outside_click and (m_selection_type= e_unknown_selection)
                      then begin
                          draw_selecter;
                          empty_selecter;
                          m_mouse_x:= 0;+ m_mouse_y:= 0;
                        end;
                  end;
            end;
      end; // f_handled_mouse_down

    function c_rectangle_selecter.f_handled_mouse_move(p_shift_state: TShiftState; p_x, p_y: Integer): Boolean;
      var l_delta_x, l_delta_y: Integer;
      begin
        // -- has selected from caller, and has already clicked, with left button
        Result:= m_is_active and (Abs(m_mouse_x)+ Abs(m_mouse_y)<> 0)
          and ((ssLeft in p_shift_state) or (ssRight in p_shift_state))
          ;

        if Result
          then begin
              // -- save for bezier translation or scaling
              m_previous_x_1:= m_x_1;
              m_previous_y_1:= m_y_1;
              m_previous_x_2:= m_x_2;
              m_previous_y_2:= m_y_2;

              // -- first draw
              if f_is_empty
                then begin
                    // display('first');
                    m_x_1:= m_mouse_x; m_y_1:= m_mouse_y;
                    m_x_2:= p_x; m_y_2:= p_y;
                    draw_selecter;
                  end
                else begin
                    // -- delete
                    draw_selecter;

                    case m_selection_type of
                      e_south_east_selection :
                        begin
                          m_x_2:= p_x; m_y_2:= p_y;
                        end;
                      e_south_west_selection :
                        begin
                          m_x_1:= p_x; m_y_2:= p_y;
                        end;
                      e_north_west_selection :
                        begin
                          m_x_1:= p_x; m_y_1:= p_y;
                        end;
                      e_north_east_selection :
                        begin
                          m_x_2:= p_x; m_y_1:= p_y;
                        end;

                      e_north_selection :
                        begin
                          m_y_1:= p_y;
                        end;
                      e_south_selection :
                        begin
                          m_y_2:= p_y;
                        end;
                      e_east_selection :
                        begin
                          m_x_2:= p_x;
                        end;
                      e_west_selection :
                        begin
                          m_x_1:= p_x; 
                        end;

                      e_inside_selection :
                        begin
                          l_delta_x:= p_x- m_mouse_x;
                          l_delta_y:= p_y- m_mouse_y;
                          Inc(m_x_1, l_delta_x); Inc(m_x_2, l_delta_x);
                          Inc(m_y_1, l_delta_y); Inc(m_y_2, l_delta_y);
                          m_mouse_x:= p_x;
                          m_mouse_y:= p_y
                        end;
                      else
                    end; // case

                    // -- draw elsewhere
                    draw_selecter;
                  end;
            end;
      end; // f_handled_mouse_move

    function c_rectangle_selecter.f_handled_mouse_up(p_mouse_button: TMouseButton;
        p_shift_state: TShiftState; p_x, p_y: Integer): Boolean;
      begin
        Result:= m_is_active;
      end; // f_handled_mouse_up

    function c_rectangle_selecter.f_is_in_selecter(p_x, p_y: Integer): Boolean;
      begin
        Result:= (p_x>= m_x_1) and (p_x< m_x_2)
            and (p_y>= m_y_1) and (p_y< m_y_2)
      end; // f_is_in_selecter

    procedure c_rectangle_selecter.move_point(var pv_x, pv_y: Double);
      begin
        if (m_x_2<= m_x_1) or (m_y_2<= m_y_1)
          then Exit;

        // -- stretch by rule of thumb wr to the opposite corner
        case m_selection_type of
          e_south_east_selection :
            begin
              pv_x:= pv_x+ (pv_x- m_x_1)* (m_x_2- m_previous_x_2)/ (m_x_2- m_x_1);
              pv_y:= pv_y+ (pv_y- m_y_1)* (m_y_2- m_previous_y_2)/ (m_y_2- m_y_1);
            end;
          e_south_west_selection :
            begin
              pv_x:= pv_x- (pv_x- m_x_2)* (m_x_1- m_previous_x_1)/ (m_x_2- m_x_1);
              pv_y:= pv_y+ (pv_y- m_y_1)* (m_y_2- m_previous_y_2)/ (m_y_2- m_y_1);
            end;
          e_north_west_selection :
            begin
              pv_x:= pv_x- (pv_x- m_x_2)* (m_x_1- m_previous_x_1)/ (m_x_2- m_x_1);
              pv_y:= pv_y- (pv_y- m_y_2)* (m_y_1- m_previous_y_1)/ (m_y_2- m_y_1);
            end;
          e_north_east_selection :
            begin
              pv_x:= pv_x+ (pv_x- m_x_1)* (m_x_2- m_previous_x_2)/ (m_x_2- m_x_1);
              pv_y:= pv_y- (pv_y- m_y_2)* (m_y_1- m_previous_y_1)/ (m_y_2- m_y_1);
            end;

          e_north_selection :
            begin
              pv_y:= pv_y- (pv_y- m_y_2)* (m_y_1- m_previous_y_1)/ (m_y_2- m_y_1);
            end;
          e_south_selection :
            begin
              pv_y:= pv_y+ (pv_y- m_y_1)* (m_y_2- m_previous_y_2)/ (m_y_2- m_y_1);
            end;
          e_east_selection :
            begin
              pv_x:= pv_x+ (pv_x- m_x_1)* (m_x_2- m_previous_x_2)/ (m_x_2- m_x_1);
            end;
          e_west_selection :
            begin
              pv_x:= pv_x- (pv_x- m_x_2)* (m_x_1- m_previous_x_1)/ (m_x_2- m_x_1);
            end;

          e_inside_selection :
            begin
              pv_x:= pv_x+ m_x_1- m_previous_x_1;
              pv_y:= pv_y+ m_y_1- m_previous_y_1;
            end;
        end; // case
      end; // move_point

    begin // u_c_selecter
    end. // u_c_selecter
