// 002 u_c_bezier_transformer
// 14 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_bezier_transformer;
  interface
    uses Classes, Windows
        , u_c_basic_object
        , u_c_2d_c_point_list
        , u_c_transformer
        ;

    type c_bezier_transformer= class(c_transformer)
                                 constructor create_bezier_transformer(p_name: String);

                                 function f_c_transform(p_c_2d_point_list: c_2d_point_list;
                                     p_interpolation_count: Integer): c_2d_point_list; Override;
                                 function f_interpolate_value(p_c_2d_point_list: c_2d_point_list;
                                     p_x: Double): Double; Override;
                               end; // create_bezier_transformer

  implementation
    uses SysUtils, u_c_display, u_double
         ;

    Function f_smooth_bezier(p_parameter: Double; p_point: Byte): Double;
      Begin
        Result:=0;
        case p_point of
           0 : begin
                 p_parameter:= 1- p_parameter;
                 Result:= p_parameter* p_parameter* p_parameter/ 6
               end;
           1 : Result:= p_parameter* p_parameter* (p_parameter/ 2- 1)+ 4/6;
           2 : Result:= p_parameter* (1+ p_parameter* (1- p_parameter))/ 2+ 1/6;
           3 : Result:= p_parameter* p_parameter* p_parameter/ 6;
        End; // case p_point
      end; // f_smooth_bezier

    // -- c_bezier_transformer

    constructor c_bezier_transformer.create_bezier_transformer(p_name: String);
      begin
        inherited create_basic_object(p_name);
      end; // create_bezier_transformer

    function c_bezier_transformer.f_c_transform(p_c_2d_point_list: c_2d_point_list;
        p_interpolation_count: Integer): c_2d_point_list;
      Var l_point_index: Integer;
          l_c_start_point, l_c_end_point: c_2d_point;
          l_c_before_point, l_c_after_point: c_2d_point;
          l_bezier_parameter, l_bezier_parameter_step, l_distance: Double;
          l_x, l_y: Double;
      begin
        Result:= c_2d_point_list.create_2d_point_list('bezier');

        with p_c_2d_point_list do
          For l_point_index:= 1 to f_2d_point_count- 2- 1 do
          begin
            l_c_start_point:= f_c_2d_point(l_point_index);
            l_c_end_point:= f_c_2d_point(l_point_index+ 1);
            l_distance:= Sqrt(Sqr(l_c_end_point.m_x- l_c_start_point.m_x)
                + Sqr(l_c_end_point.m_y- l_c_start_point.m_y));
            if l_distance= 0
              then l_distance:= 1;

            // -- for Bezier, the step is a value in 0..1
            l_bezier_parameter_step:= 1 / p_interpolation_count;
            l_bezier_parameter:= 0;

            l_c_before_point:= f_c_2d_point(l_point_index- 1);
            l_c_after_point:= f_c_2d_point(l_point_index+ 2);

            // -- one or several interpolated points
            repeat
              l_x:= f_smooth_bezier(l_bezier_parameter, 0)* l_c_before_point.m_x
                  + f_smooth_bezier(l_bezier_parameter, 1)* l_c_start_point.m_x
                  + f_smooth_bezier(l_bezier_parameter, 2)* l_c_end_point.m_x
                  + f_smooth_bezier(l_bezier_parameter, 3)* l_c_after_point.m_x;
              l_y:= f_smooth_bezier(l_bezier_parameter, 0)* l_c_before_point.m_y
                  + f_smooth_bezier(l_bezier_parameter, 1)* l_c_start_point.m_y
                  + f_smooth_bezier(l_bezier_parameter, 2)* l_c_end_point.m_y
                  + f_smooth_bezier(l_bezier_parameter, 3)* l_c_after_point.m_y;

              Result.add_2d_point('', c_2d_point.create_2d_point('', l_x, l_y));

              l_bezier_parameter:= l_bezier_parameter+ l_bezier_parameter_step;
            until l_bezier_parameter> 1;
          End; // For l_point_index
      end; // f_c_transform

    function c_bezier_transformer.f_interpolate_value(p_c_2d_point_list: c_2d_point_list;
        p_x: Double): Double;
        // -- assumes a FUNCTION (a single y value for each x)

      function f_interpolate(p_point_index: Integer): Double;
          // -- interpolate
        var l_c_start_point, l_c_end_point: c_2d_point;
            l_c_before_point, l_c_after_point: c_2d_point;
            l_bezier_parameter, l_bezier_parameter_step: Double;
            l_distance, l_x: Double;
        begin
          with p_c_2d_point_list do
          begin
            l_c_before_point:= f_c_2d_point(p_point_index- 1);
            l_c_start_point:= f_c_2d_point(p_point_index);
            l_c_end_point:= f_c_2d_point(p_point_index+ 1);
            l_c_after_point:= f_c_2d_point(p_point_index+ 2);

            l_distance:= Sqrt(Sqr(l_c_end_point.m_x- l_c_start_point.m_x)
                + Sqr(l_c_end_point.m_y- l_c_start_point.m_y));
            if l_distance= 0
              then l_distance:= 1;
            // -- at least a value for each integer x
            // --   (for horizontal should be 1 / distance)
            l_bezier_parameter_step:= 1/ l_distance/ 2;

            l_bezier_parameter:= - l_bezier_parameter_step;
            repeat
              l_bezier_parameter:= l_bezier_parameter+ l_bezier_parameter_step;

              l_x:= f_smooth_bezier(l_bezier_parameter, 0)* l_c_before_point.m_x
                  + f_smooth_bezier(l_bezier_parameter, 1)* l_c_start_point.m_x
                  + f_smooth_bezier(l_bezier_parameter, 2)* l_c_end_point.m_x
                  + f_smooth_bezier(l_bezier_parameter, 3)* l_c_after_point.m_x;
            until (l_bezier_parameter_step> 1) or (l_x> p_x);

            Result:= f_smooth_bezier(l_bezier_parameter, 0)* l_c_before_point.m_y
                + f_smooth_bezier(l_bezier_parameter, 1)* l_c_start_point.m_y
                + f_smooth_bezier(l_bezier_parameter, 2)* l_c_end_point.m_y
                + f_smooth_bezier(l_bezier_parameter, 3)* l_c_after_point.m_y;
          end; // with p_c_2d_point_list
        end; // f_interpolate

      var l_point_index: Integer;

      begin // f_bezier_value
        // -- find the interval
        with p_c_2d_point_list do
        begin
          if f_2d_point_count= 0
            then begin
                Result:= 0;
                Exit;
              end;

          l_point_index:= 0;
          while l_point_index<= f_2d_point_count- 1 do
            if f_c_2d_point(l_point_index).m_x> p_x
              then Break
              else Inc(l_point_index);

          if l_point_index<= 1
            then begin
                Result:= f_c_2d_point(0).m_y;
              end
            else
              if l_point_index> f_2d_point_count- 2
                then begin
                    l_point_index:= f_2d_point_count- 1;
                    Result:= f_c_2d_point(l_point_index).m_y;
                  end
                else Result:= f_interpolate(l_point_index- 1);
          // display(f_display_2d_point(f_c_2d_point(l_point_index)));
        end; // with p_c_2d_point_list
      end; // f_interpolate_value

  end.
