// 002 u_c_poly_line_transformer
// 14 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_poly_line_transformer;
  interface
    uses Classes, Windows
        , u_c_basic_object
        , u_c_2d_c_point_list
        , u_c_transformer
        ;

    type c_poly_line_transformer= class(c_transformer)
                                    constructor create_poly_line_transformer(p_name: String);

                                    function f_c_transform(p_c_2d_point_list: c_2d_point_list;
                                        p_interpolation_count: Integer): c_2d_point_list; Override;
                                    function f_interpolate_value(p_c_2d_point_list: c_2d_point_list;
                                        p_x: Double): Double; Override;
                                  end; // c_poly_line_transformer

  implementation
    uses SysUtils, u_c_display, u_double
         ;

    // -- c_poly_line_transform

    constructor c_poly_line_transformer.create_poly_line_transformer(p_name: String);
      begin
        inherited create_basic_object(p_name);
      end; // create_poly_line_transformer

    function c_poly_line_transformer.f_c_transform(p_c_2d_point_list: c_2d_point_list;
        p_interpolation_count: Integer): c_2d_point_list;
      Var l_point_index: Integer;
          l_c_start_point, l_c_end_point: c_2d_point;
          l_poly_parameter, l_poly_parameter_step, l_distance: Double;

          l_delta_x, l_delta_y: Double;
          l_slope: Double;
          l_x, l_y: Double;
      begin
        Result:= c_2d_point_list.create_2d_point_list('bezier');

        with p_c_2d_point_list do
          For l_point_index:= 0 to f_2d_point_count- 2 do
          begin
            l_c_start_point:= f_c_2d_point(l_point_index);
            l_c_end_point:= f_c_2d_point(l_point_index+ 1);

            l_distance:= Sqrt(Sqr(l_c_end_point.m_x- l_c_start_point.m_x)
                + Sqr(l_c_end_point.m_y- l_c_start_point.m_y));

            if l_distance= 0
              then
              else begin
                  l_poly_parameter_step:= l_distance / p_interpolation_count;

                  l_poly_parameter:= 0;
                  // display('dist '+ f_display_double(l_distance, 3, 2)
                  //   + ' param '+ f_display_double(l_poly_parameter_step, 3, 2));

                  l_delta_x:= l_c_end_point.m_x- l_c_start_point.m_x;
                  l_delta_y:= l_c_end_point.m_y- l_c_start_point.m_y;

                  if f_double_is_null(l_delta_x)
                    then begin
                        // display('same_x');
                        // -- one or several interpolated points
                        repeat
                          l_x:= l_c_start_point.m_x;
                          l_y:= l_c_start_point.m_y+ l_poly_parameter_step;

                          Result.add_2d_point('', c_2d_point.create_2d_point('', l_x, l_y));

                          l_poly_parameter:= l_poly_parameter+ l_poly_parameter_step;
                        until l_y+ l_poly_parameter> l_c_end_point.m_y;

                        // -- add the last segment
                        // -- todo
                      end
                    else begin
                        l_slope:= l_delta_y / l_delta_x ;
                        // display('step '+ f_display_double(l_poly_parameter_step, 3, 2));

                        // -- one or several interpolated points
                        repeat
                          l_x:= l_c_start_point.m_x+ l_poly_parameter;
                          l_y:= l_c_start_point.m_y+ l_slope* (l_x- l_c_start_point.m_x);

                          // display('  add');
                          Result.add_2d_point('', c_2d_point.create_2d_point('', l_x, l_y));

                          l_poly_parameter:= l_poly_parameter+ l_poly_parameter_step;
                        until l_x+ l_poly_parameter> l_c_end_point.m_x;

                        // -- add the last segment
                        if l_point_index= f_2d_point_count- 2
                          then
                            with f_c_2d_point(f_2d_point_count- 1) do
                              Result.add_2d_point('', c_2d_point.create_2d_point('', m_x, m_y));
                      end;
                end;
          End; // For l_point_index
      end; // f_c_transform

    function c_poly_line_transformer.f_interpolate_value(p_c_2d_point_list: c_2d_point_list;
        p_x: Double): Double;
        // -- for each p_x, find the y value

      function f_interpolate(p_point_index: Integer): Double;
          // -- interpolate
        Var l_c_start_point, l_c_end_point: c_2d_point;
            l_delta_x, l_delta_y: Double;
            l_slope: Double;
            l_x, l_y: Double;
        begin
          with p_c_2d_point_list do
          begin
            l_c_start_point:= f_c_2d_point(p_point_index);
            l_c_end_point:= f_c_2d_point(p_point_index+ 1);

            l_delta_x:= l_c_end_point.m_x- l_c_start_point.m_x;
            l_delta_y:= l_c_end_point.m_y- l_c_start_point.m_y;

            if f_double_is_null(l_delta_x)
              then begin
                  Result:= l_c_start_point.m_y;
                end
              else begin
                  l_slope:= l_delta_y / l_delta_x ;

                  Result:= l_c_start_point.m_y+ l_slope* (p_x- l_c_start_point.m_x);
                end;
          end; // with p_c_2d_point_list
        end; // f_interpolate

      var l_point_index: Integer;
        
      begin // f_interpolate_value
        // -- find the interval
        with p_c_2d_point_list do
        begin
          if f_2d_point_count= 0
            then begin
                Result:= 0;
                Exit;
              end
            else
              if f_2d_point_count= 1
                then begin
                    Result:= f_c_2d_point(0).m_y;
                    Exit;
                  end;

          l_point_index:= 0;
          while l_point_index<= f_2d_point_count- 1 do
            if f_c_2d_point(l_point_index).m_x> p_x
              then Break
              else Inc(l_point_index);


          if l_point_index< 1
            then begin
                // -- before the first point
                Result:= f_c_2d_point(0).m_y;
              end
            else
              if l_point_index> f_2d_point_count- 1
                then begin
                    // -- after the last point
                    l_point_index:= f_2d_point_count- 1;
                    Result:= f_c_2d_point(l_point_index).m_y;
                  end
                else Result:= f_interpolate(l_point_index- 1);
        end; // with p_c_2d_point_list
      end; // f_interpolate_value

  end.
