// 002 u_c_bitmap_direct
// 15 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

// -- save the original rgb_float, hsv_float,
// -- transform those into "running" rgb and hsv
// -- display by converting into Scanline

(*$r+*)

unit u_c_bitmap_direct;
  interface
    uses Graphics
        , u_c_basic_object
        , u_graphic
        , u_c_histogram_255
        ;

    type t_oa_rgb_float_array= array of t_RGB_float_array;
         t_oa_hsv_float_array= array of t_hsv_float_array;

         c_bitmap= class(c_basic_object)
                     m_c_bitmap: tBitMap;
                     m_c_canvas: tCanvas;

                     m_line_count, m_column_count: Integer;
                     m_pixel_count: Integer;

                     // -- our computation array: lines of RGB arrays
                     m_oa_rgb_lines, m_oa_original_rgb_lines: t_oa_rgb_float_array;

                     m_selecter_x_min, m_selecter_y_min, m_selecter_x_max, m_selecter_y_max: Integer;
                     m_c_red_histogram, m_c_green_histogram, m_c_blue_histogram: c_histogram_255;

                     // -- work in hsv domain
                     m_oa_hsv_lines, m_oa_original_hsv_lines: t_oa_hsv_float_array;
                     m_c_color_histogram, m_c_saturation_histogram, m_c_intensity_histogram: c_histogram_255;
                     // -- at this level for Form1.display
                     m_saturation_min, m_saturation_max: Double;
                     m_intensity_min, m_intensity_max: Double;

                     Constructor create_bitmap(p_name: String; p_c_canvas: tCanvas);

                     procedure display_bitmap_info;
                     procedure draw_bitmap;

                     procedure create_lines;
                     procedure copy_bitmap_to_rgb_lines(const pk_rgb_lines: t_oa_rgb_float_array);
                     procedure copy_rgb_lines_to_bitmap;
                     procedure copy_orignal_rgb_lines_to_rgb_lines;

                       procedure set_rgb_color(p_delta_rgb: Double; p_contrast: Boolean;
                           p_rgb_color_type: t_rgb_color_type);
                     procedure set_rgb_red(p_red_value: Double; p_contrast: Boolean);
                     procedure set_rgb_green(p_green_value: Double; p_contrast: Boolean);
                     procedure set_rgb_blue(p_blue_value: Double; p_contrast: Boolean);

                     // -- helpers
                     procedure compute_y_and_y_start_end(var pv_start_x, pv_end_x, pv_start_y, pv_end_y: integer);
                     procedure set_selecter_min_max(p_selecter_x_min, p_selecter_y_min, p_selecter_x_max, p_selecter_y_max: Integer);

                     procedure copy_bitmap_to_hsv_lines(const pk_oa_hsv_float_array: t_oa_hsv_float_array);
                     procedure convert_hsv_lines_to_rgb_lines;
                     procedure copy_original_hsv_lines_to_hsv_lines;

                     procedure rotate_hsv_color(p_delta_color: Double);
                     procedure change_hsv_saturation(p_delta_saturation: Double; p_contrast: Boolean);
                     procedure change_hsv_intensity(p_delta_intensity: Double; p_contrast: Boolean);

                     procedure create_color_histograms(p_c_canvas: tCanvas);
                     procedure compute_color_histograms;
                     procedure rescale_and_draw_color_histograms(p_y_max, p_fat: Integer);

                     procedure create_hsv_histograms(p_c_canvas: tCanvas);
                     procedure compute_hsv_histograms;
                     procedure rescale_and_draw_hsv_histograms(p_y_max, p_fat: Integer);

                     Destructor Destroy; Override;
                   end; // c_bitmap

  implementation
    uses SysUtils, Math
        , u_c_display
        , u_double
       ;

    // -- c_bitmap

    Constructor c_bitmap.create_bitmap(p_name: String; p_c_canvas: tCanvas);
      begin
        Inherited create_basic_object(p_name);
        m_c_canvas:= p_c_canvas;
      end; // create_bitmap

    procedure c_bitmap.display_bitmap_info;
      begin
        with m_c_bitmap do
        begin
          display('Width   '+ IntToStr(Width));
          display('Height  '+ IntToStr(Height));
        end;
      end; // display_bitmap_info

    procedure c_bitmap.draw_bitmap;
      begin
        m_c_canvas.Draw(0, 0, m_c_bitmap);
      end; // draw_bitmap

    // -- initialize the float bitmaps

    procedure c_bitmap.create_lines;
      var l_line_index: Integer;
      begin
        m_line_count:= m_c_bitmap.Height;
        m_column_count:= m_c_bitmap.Width;
        m_pixel_count:= m_line_count* m_column_count;

        // -- create the 4 bitmaps
        SetLength(m_oa_rgb_lines, m_line_count);
        for l_line_index:= 0 to m_line_count- 1 do
          SetLength(m_oa_rgb_lines[l_line_index], m_column_count);
        SetLength(m_oa_original_rgb_lines, m_line_count);
        for l_line_index:= 0 to m_line_count- 1 do
          SetLength(m_oa_original_rgb_lines[l_line_index], m_column_count);

        SetLength(m_oa_hsv_lines, m_line_count);
        for l_line_index:= 0 to m_line_count- 1 do
          SetLength(m_oa_hsv_lines[l_line_index], m_column_count);
        SetLength(m_oa_original_hsv_lines, m_line_count);
        for l_line_index:= 0 to m_line_count- 1 do
          SetLength(m_oa_original_hsv_lines[l_line_index], m_column_count);

        // -- copy the tBitmap to the float bitmaps
        copy_bitmap_to_rgb_lines(m_oa_original_rgb_lines);
        copy_bitmap_to_rgb_lines(m_oa_rgb_lines);
        // -- copy the tBitmap to the rgb float bitmaps
        copy_bitmap_to_hsv_lines(m_oa_original_hsv_lines);
        copy_bitmap_to_hsv_lines(m_oa_hsv_lines);
      end; // create_lines

    procedure c_bitmap.copy_bitmap_to_rgb_lines(const pk_rgb_lines: t_oa_rgb_float_array);
      var l_line_index, l_column_index: Integer;
          l_pt_source_row: t_pt_RGB_array;
          l_red, l_green, l_blue: Integer;
      begin
        For l_line_index:= 0 to m_line_count- 1 do
        begin
          l_pt_source_row:= m_c_bitmap.Scanline[l_line_index];

          for l_column_index:= 0 to m_column_count- 1 do
          begin
            (*$r-*)
            l_red:= l_pt_source_row[l_column_index].m_red;
            l_green:= l_pt_source_row[l_column_index].m_green;
            l_blue:= l_pt_source_row[l_column_index].m_blue;
            (*$r+*)

            with pk_rgb_lines[l_line_index][l_column_index] do
            begin
              m_float_red:= l_red;
              m_float_green:= l_green;
              m_float_blue:= l_blue;
            end;
          end;
        end; // for l_line
      end; // copy_bitmap_to_rgb_lines

    procedure c_bitmap.copy_rgb_lines_to_bitmap;
      var l_line_index, l_column_index: Integer;
          l_pt_source_row: t_pt_RGB_array;
          l_red, l_green, l_blue: t_float;
      begin
        For l_line_index:= 0 to m_line_count- 1 do
        begin
          l_pt_source_row:= m_c_bitmap.Scanline[l_line_index];

          for l_column_index:= 0 to m_column_count- 1 do
          begin
            (*$r-*)
            l_red:= m_oa_rgb_lines[l_line_index][l_column_index].m_float_red;
            l_pt_source_row[l_column_index].m_red:= f_float_to_byte(l_red);

            l_green:= m_oa_rgb_lines[l_line_index][l_column_index].m_float_green;
            l_pt_source_row[l_column_index].m_green:= f_float_to_byte(l_green);

            l_blue:= m_oa_rgb_lines[l_line_index][l_column_index].m_float_blue;
            l_pt_source_row[l_column_index].m_blue:= f_float_to_byte(l_blue);
            (*$r+*)
          end;
        end; // for l_line
      end; // copy_rgb_lines_to_bitmap

    procedure c_bitmap.copy_orignal_rgb_lines_to_rgb_lines;
      var l_line_index, l_column_index: Integer;
      begin
        for l_line_index:= 0 to m_line_count- 1 do
          for l_column_index:= 0 to m_column_count- 1 do
            m_oa_rgb_lines[l_line_index][l_column_index]:= m_oa_original_rgb_lines[l_line_index][l_column_index];
      end; // copy_orignal_rgb_lines_to_rgb_lines

    // -- work on rgb

    procedure c_bitmap.set_rgb_color(p_delta_rgb: Double; p_contrast: Boolean;
        p_rgb_color_type: t_rgb_color_type);
      var l_line_index, l_column_index: Integer;
          l_contrast_factor: t_float;
      begin
        if p_contrast
          then begin
              // -- 0 to 50, 10=no change
              // -- < 10 bowl, > 10 dome
              l_contrast_factor:= 256/ Power(256, p_delta_rgb);
            end
          else // -- avoid warnings
               l_contrast_factor:= 1;

        for l_line_index:= 0 to m_line_count- 1 do
          for l_column_index:= 0 to m_column_count- 1 do
            with m_oa_rgb_lines[l_line_index][l_column_index] do
              case p_rgb_color_type of
                e_red_color :
                  if p_contrast
                    then m_float_red:= Power(m_float_red, p_delta_rgb)* l_contrast_factor
                    else m_float_red:= m_float_red+ p_delta_rgb;
                e_green_color :
                  if p_contrast
                    then m_float_green:= Power(m_float_green, p_delta_rgb)* l_contrast_factor
                    else m_float_green:= m_float_green+ p_delta_rgb;
                e_blue_color :
                  if p_contrast
                    then m_float_blue:= Power(m_float_blue, p_delta_rgb)* l_contrast_factor
                    else m_float_blue:= m_float_blue+ p_delta_rgb;
              end; // case p_rgb_color_type
      end; // set_rgb_color

    procedure c_bitmap.set_rgb_red(p_red_value: Double; p_contrast: Boolean);
      begin
        set_rgb_color(p_red_value, p_contrast, e_red_color);
(*
      var l_line_index, l_column_index: Integer;
          l_contrast_factor: t_float;

        if p_contrast
          then begin
              // -- 0 to 50, 10=no change
              // -- < 10 bowl, > 10 dome
              l_contrast_factor:= 256/ Power(256, p_red_value);
            end;

        for l_line_index:= 0 to m_line_count- 1 do
          for l_column_index:= 0 to m_column_count- 1 do
            with m_oa_rgb_lines[l_line_index][l_column_index] do
              if p_contrast
                then begin
                    m_float_red:= Power(m_float_red, p_red_value)* l_contrast_factor;
                  end
                else m_float_red:= m_float_red+ p_red_value;
*)
      end; // set_rgb_red

    procedure c_bitmap.set_rgb_green(p_green_value: Double; p_contrast: Boolean);
      begin
        set_rgb_color(p_green_value, p_contrast, e_green_color);
      end; // set_rgb_green

    procedure c_bitmap.set_rgb_blue(p_blue_value: Double; p_contrast: Boolean);
      begin
        set_rgb_color(p_blue_value, p_contrast, e_blue_color);
      end; // set_rgb_blue

    // -- work on hsv

    procedure c_bitmap.copy_bitmap_to_hsv_lines(const pk_oa_hsv_float_array: t_oa_hsv_float_array);
        // -- do it on the original point
      var l_line_index, l_column_index: Integer;
          l_pt_source_row: t_pt_RGB_array;
          l_rgb_float_triple: t_rgb_float_triple;
      begin
        For l_line_index:= 0 to m_line_count- 1 do
        begin
          l_pt_source_row:= m_c_bitmap.Scanline[l_line_index];

          for l_column_index:= 0 to m_column_count- 1 do
          begin
            (*$r-*)
            with l_pt_source_row[l_column_index], l_rgb_float_triple do
            (*$r+*)
            begin
              m_float_red:= m_red; m_float_green:= m_green; m_float_blue:= m_blue;
            end;

            pk_oa_hsv_float_array[l_line_index, l_column_index]:= f_rgb_to_hsv(l_rgb_float_triple);
          end; // for l_column_index
        end; // for l_line_index
      end; // copy_bitmap_to_hsv_lines

    procedure c_bitmap.rotate_hsv_color(p_delta_color: Double);
        // -- rotate the working array
      var l_line_index, l_column_index: Integer;
          l_color: Double;
      begin
        For l_line_index:= 0 to m_line_count- 1 do
          for l_column_index:= 0 to m_column_count- 1 do
            with m_oa_hsv_lines[l_line_index, l_column_index] do
            begin
              l_color:= m_color+ p_delta_color;
              if l_color< 0
                then l_color:= l_color+ 360;
              if l_color> 360
                then l_color:= l_color- (l_color / 360)* 360;
              m_color:= l_color;    
            end;
      end; // rotate_hsv_color

    procedure c_bitmap.change_hsv_saturation(p_delta_saturation: Double; p_contrast: Boolean);
      var l_line_index, l_column_index: Integer;
          l_contrast_factor: t_float;
      begin
        if p_contrast
          then begin
              // l_contrast_factor:= 256/ Power(256, p_delta_saturation);
              l_contrast_factor:= Power(256, p_delta_saturation)/ 256;
            end
          else // -- avoid warning
               l_contrast_factor:= 1.0 ;

        For l_line_index:= 0 to m_line_count- 1 do
          for l_column_index:= 0 to m_column_count- 1 do
            with m_oa_hsv_lines[l_line_index, l_column_index] do
              if p_contrast
                then m_saturation:= Power(m_saturation, l_contrast_factor) // * l_contrast_factor
                else m_saturation:= m_saturation+ p_delta_saturation;
      end; // change_hsv_saturation

    procedure c_bitmap.change_hsv_intensity(p_delta_intensity: Double; p_contrast: Boolean);
      var l_line_index, l_column_index: Integer;
          l_contrast_factor: t_float;
      begin
        if p_contrast
          then begin
              l_contrast_factor:= Power(256, p_delta_intensity)/ 256/ 2;
            end
          else // -- avoid warning
               l_contrast_factor:= 1.0 ;

        for l_line_index:= 0 to m_line_count- 1 do
          for l_column_index:= 0 to m_column_count- 1 do
            with m_oa_hsv_lines[l_line_index, l_column_index] do
              if p_contrast
                then // m_intensity:= Power(m_intensity, p_delta_intensity* 5) // * l_contrast_factor
                     m_intensity:= Power(m_intensity, l_contrast_factor) // * l_contrast_factor
                else m_intensity:= m_intensity+ p_delta_intensity;
      end; // change_hsv_intensity

    procedure c_bitmap.convert_hsv_lines_to_rgb_lines;
        // -- convert back the hsv array into the float array
        // -- the HSV values might be illegal (color in 0..360, saturation in 0..1, intensity in 0..1
      var l_line_index, l_column_index: Integer;
          l_hsv: t_hsv_triple;
          l_rgb: t_rgb_float_triple;
      begin
        For l_line_index:= 0 to m_line_count- 1 do
          for l_column_index:= 0 to m_column_count- 1 do
          begin
            // l_hsv:= f_normalize_hsv(m_oa_hsv_lines[l_line_index, l_column_index]);
            l_hsv:= m_oa_hsv_lines[l_line_index, l_column_index];
            l_rgb:= f_hsv_to_rgb(l_hsv);

            with l_rgb do
            begin
              m_float_red:= m_float_red* 255.0;
              m_float_green:= m_float_green* 255.0;
              m_float_blue:= m_float_blue* 255.0;
            end;
            m_oa_rgb_lines[l_line_index, l_column_index]:= l_rgb;
          end;
      end; // convert_hsv_lines_to_rgb_lines

    procedure c_bitmap.copy_original_hsv_lines_to_hsv_lines;
      var l_line_index, l_column_index: Integer;
      begin
        For l_line_index:= 0 to m_line_count- 1 do
          for l_column_index:= 0 to m_column_count- 1 do
            m_oa_hsv_lines[l_line_index, l_column_index]:= m_oa_original_hsv_lines[l_line_index, l_column_index];
      end; // copy_original_hsv_lines_to_hsv_lines

    // -- histograms

    procedure c_bitmap.set_selecter_min_max(p_selecter_x_min, p_selecter_y_min, p_selecter_x_max, p_selecter_y_max: Integer);
      begin
        m_selecter_x_min:= p_selecter_x_min;
        m_selecter_x_max:= p_selecter_x_max;
        m_selecter_y_min:= p_selecter_y_min;
        m_selecter_y_max:= p_selecter_y_max;
      end; // set_selecter_min_max

    procedure c_bitmap.compute_y_and_y_start_end(var pv_start_x, pv_end_x, pv_start_y, pv_end_y: integer);
        // -- only used for histograms
      begin
(*
        pv_start_x:= 0;
        pv_start_y:= 0;
        pv_end_x:= m_c_bitmap.Width- 1;
        pv_end_y:= m_c_bitmap.Height- 1;
*)
        pv_start_x:= m_selecter_x_min;
        pv_start_y:= m_selecter_y_min;
        pv_end_x:= m_selecter_x_max;
        pv_end_y:= m_selecter_y_max;
      end; // compute_y_and_y_start_end

    // -- histograms

    procedure c_bitmap.create_color_histograms(p_c_canvas: tCanvas);
      begin
        m_c_red_histogram.Free; m_c_green_histogram.Free; m_c_blue_histogram.Free;

        m_c_red_histogram:= c_histogram_255.create_histogram_255('red', p_c_canvas, clRed);
        m_c_green_histogram:= c_histogram_255.create_histogram_255('green', p_c_canvas, clLime);
        m_c_blue_histogram:= c_histogram_255.create_histogram_255('blue', p_c_canvas, clBlue);
      end; // create_color_histograms

    procedure c_bitmap.compute_color_histograms;
      var l_start_x, l_end_x, l_start_y, l_end_y: integer;
          l_x, l_y: integer;
          l_pt_source_row: t_pt_RGB_array;
      begin
        compute_y_and_y_start_end(l_start_x, l_end_x, l_start_y, l_end_y);
        // display(Format('%3d..%3d %3d..%3d', [l_start_x, l_end_x, l_start_y, l_end_y]));

        m_c_red_histogram.reset_histogram;
        m_c_green_histogram.reset_histogram;
        m_c_blue_histogram.reset_histogram;

        For l_y:= l_start_y to l_end_y do
        begin
          l_pt_source_row:= m_c_bitmap.Scanline[l_y];

          // -- Fetch the colors
          for l_x:= l_start_x to l_end_x do
            (*$r-*)
            with l_pt_source_row^[l_x] do
            (*$r+*)
            begin
              Inc(m_c_red_histogram.m_histogram[m_red]);
              Inc(m_c_green_histogram.m_histogram[m_green]);
              Inc(m_c_blue_histogram.m_histogram[m_blue]);
            end;
        end; // For l_y
      end; // compute_color_histograms

    procedure c_bitmap.rescale_and_draw_color_histograms(p_y_max, p_fat: Integer);
      var l_total_max: Integer;
          l_drawing_height: Integer;
          l_y: Integer;

      procedure do_rescale_and_draw(const pk_c_histogram_255: c_histogram_255);
        var l_the_height: Integer;
        begin
          with pk_c_histogram_255 do
          begin
            l_the_height:= l_drawing_height* m_max div l_total_max;
            rescale_histogram(l_the_height);

            draw_histogram_with_border(p_y_max- l_y, p_fat);

            Inc(l_y, l_the_height+ 2* p_fat);
          end; // with m_c_red_histogram
        end; // do_rescale_and_draw

      begin // rescale_and_draw_color_histograms
        // display(Format('draw height %5d fat=%5d', [p_y_max, p_fat]));
        m_c_red_histogram.compute_max;
        m_c_green_histogram.compute_max;
        m_c_blue_histogram.compute_max;

        l_total_max:= m_c_red_histogram.m_max+ m_c_green_histogram.m_max
            + m_c_blue_histogram.m_max;

        l_drawing_height:= p_y_max- 6* p_fat;
        l_y:= 0;

        do_rescale_and_draw(m_c_blue_histogram);
        do_rescale_and_draw(m_c_green_histogram);
        do_rescale_and_draw(m_c_red_histogram);
      end; // rescale_color_histograms

    // -- hsv histograms

    procedure c_bitmap.create_hsv_histograms(p_c_canvas: tCanvas);
      begin
        m_c_color_histogram.Free; m_c_saturation_histogram.Free; m_c_intensity_histogram.Free;

        m_c_color_histogram:= c_histogram_255.create_histogram_255('color', p_c_canvas, clBlack);
        m_c_saturation_histogram:= c_histogram_255.create_histogram_255('saturation', p_c_canvas, clBlack);
        m_c_intensity_histogram:= c_histogram_255.create_histogram_255('intensity', p_c_canvas, clBlack);
      end; // create_hsv_histograms

    procedure c_bitmap.compute_hsv_histograms;

      function f_bin(p_value, p_value_min, p_value_max: Double): Byte;
          // -- a double between 0 and p_value max:
          // -- compute its bin between 0 and 255
        var l_spread: Double;
            l_bin: Double;
            // l_integer_bin: Integer;
        begin
          if p_value_max= 0
            then begin
                Result:= 0;
                // display('max_0');
              end
            else begin
                l_spread:= p_value_max- p_value_min;

                if l_spread= 0
                  then begin
                      Result:= 0;
                      // display('spread_0');
                    end
                  else begin
                      l_bin:= 255.0* (p_value- p_value_min)/ l_spread;
                      if l_bin> 255
                        then Result:= 255
                        else
                          if l_bin< 0
                            then begin
                                Result:= 0;
                                // display('<0');
                              end
                            else Result:= Round(l_bin);
                    end;
              end;
        end; // f_bin

      var l_start_x, l_end_x, l_start_y, l_end_y: integer;
          l_x, l_y: integer;

          l_color_min, l_color_max: Double;

          l_saturation_min, l_saturation_max: Double;
          l_intensity_min, l_intensity_max: Double;

      begin // compute_hsv_histograms
        compute_y_and_y_start_end(l_start_x, l_end_x, l_start_y, l_end_y);
        // display(Format('%3d..%3d %3d..%3d', [l_start_x, l_end_x, l_start_y, l_end_y]));

        m_c_color_histogram.reset_histogram;
        m_c_saturation_histogram.reset_histogram;
        m_c_intensity_histogram.reset_histogram;

        // -- compute the max
        m_saturation_min:= 1e10; m_saturation_max:= -1e10;
        m_intensity_min:= 1e10; m_intensity_max:= -1e10;

        For l_y:= l_start_y to l_end_y do
          for l_x:= l_start_x to l_end_x do
            with m_oa_hsv_lines[l_y, l_y] do
            begin
              m_saturation_min:= Min(m_saturation_min, m_saturation);
              m_saturation_max:= Max(m_saturation_max, m_saturation);

              m_intensity_min:= Min(m_intensity_min, m_intensity);
              m_intensity_max:= Max(m_intensity_max, m_intensity);
            end; // for l_x

        // display('col '+ f_display_double(l_color_min, 4, 0)+ '..'+ f_display_double(l_color_max, 4, 0));
        // -- always draw the colors between 0 and 355
        l_color_min:= 0;
        l_color_max:= 360;

        l_saturation_min:= m_saturation_min;
        l_saturation_max:= m_saturation_max;

        // -- correct by always including 0..1
        if l_saturation_min> 0
          then // -- start the bins at value 0
               l_saturation_min:= 0;
        if l_saturation_max< 1
          then // -- force bins up to 1.0
               l_saturation_max:= 1;

        // -- set the 0..1 displays
        m_c_saturation_histogram.m_x_1:= k_no_x;
        m_c_saturation_histogram.m_x_2:= k_no_x;
        if l_saturation_min< 0
          then // -- display the 0
               m_c_saturation_histogram.m_x_1:= f_bin(0, l_saturation_min, l_saturation_max);
        if l_saturation_max> 1
          then // -- display the value of 1.0
               m_c_saturation_histogram.m_x_2:= f_bin(1.0, l_saturation_min, l_saturation_max);

        l_intensity_min:= m_intensity_min;
        l_intensity_max:= m_intensity_max;
        // -- correct by always including 0..1
        if l_intensity_min> 0
          then // -- start the bins at value 0
               l_intensity_min:= 0;
        if l_intensity_max< 1
          then // -- force bins up to 1.0
               l_intensity_max:= 1;

        // -- set the 0..1 displays
        m_c_intensity_histogram.m_x_1:= k_no_x;
        m_c_intensity_histogram.m_x_2:= k_no_x;
        if l_intensity_min< 0
          then // -- display the 0
               m_c_intensity_histogram.m_x_1:= f_bin(0, l_intensity_min, l_intensity_max);
        if l_intensity_max> 1
          then // -- display the value of 1.0
               m_c_intensity_histogram.m_x_2:= f_bin(1.0, l_intensity_min, l_intensity_max);

        // -- now put into the bins
        For l_y:= l_start_y to l_end_y do
          for l_x:= l_start_x to l_end_x do
            with m_oa_hsv_lines[l_y, l_y] do
            begin
              Inc(m_c_color_histogram.m_histogram[f_bin(m_color, l_color_min, l_color_max)]);
              Inc(m_c_saturation_histogram.m_histogram[f_bin(m_saturation, l_saturation_min, l_saturation_max)]);
              Inc(m_c_intensity_histogram.m_histogram[f_bin(m_intensity, l_intensity_min, l_intensity_max)]);
            end; // for l_x
      end; // compute_hsv_histograms

    procedure c_bitmap.rescale_and_draw_hsv_histograms(p_y_max, p_fat: Integer);
      var l_total_max: Integer;
          l_drawing_height: Integer;
          l_y: Integer;

      procedure do_rescale_and_draw(const pk_c_histogram_255: c_histogram_255);
        var l_the_height: Integer;
        begin
          with pk_c_histogram_255 do
          begin
            l_the_height:= l_drawing_height* m_max div l_total_max;
            rescale_histogram(l_the_height);

            draw_histogram_with_border(p_y_max- l_y, p_fat);

            Inc(l_y, l_the_height+ 2* p_fat);
          end; // with m_c_red_histogram
        end; // do_rescale_and_draw

      begin // rescale_and_draw_hsv_histograms
        // display(Format('draw height %5d fat=%5d', [p_y_max, p_fat]));
        m_c_color_histogram.compute_max;
        m_c_saturation_histogram.compute_max;
        m_c_intensity_histogram.compute_max;

        l_total_max:= m_c_color_histogram.m_max+ m_c_saturation_histogram.m_max
            + m_c_intensity_histogram.m_max;

        l_drawing_height:= p_y_max- 6* p_fat;
        l_y:= 0;

        // -- draw in reverse vertical order
        do_rescale_and_draw(m_c_intensity_histogram);
        do_rescale_and_draw(m_c_saturation_histogram);
        do_rescale_and_draw(m_c_color_histogram);
      end; // rescale_and_draw_hsv_histograms

    // -- free

    Destructor c_bitmap.Destroy;
      begin
        m_c_bitmap.Free;

        m_c_red_histogram.Free; m_c_green_histogram.Free; m_c_green_histogram.Free;
        m_c_color_histogram.Free; m_c_saturation_histogram.Free; m_c_intensity_histogram.Free;

        Inherited;
      end; // Destroy

    begin // u_c_bitmap
    end. // u_c_bitmap


