// 002 u_c_histogram_255
// 14 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

// -- draw the histogram with 256 x values

(*$r+*)

unit u_c_histogram_255;
  interface
    uses Graphics, u_c_basic_object;

    const k_no_x= -12345;
    type c_histogram_255= class(c_basic_object)
                            m_histogram: array[0..255] of Integer;
                            m_c_canvas: tCanvas;
                            m_pen_color: Integer;
                            m_max: Integer;
                            // -- draw min max limits
                            m_x_1, m_x_2: Integer;

                            Constructor create_histogram_255(p_name: String;
                                p_c_canvas: tCanvas; p_pen_color: Integer);

                            procedure reset_histogram;
                            procedure display_histogram;
                            function f_total_count: Integer;
                            procedure compute_max;
                            procedure rescale_histogram(p_y_max: Integer);
                            procedure draw_histogram(p_y_min, p_y_max: Integer);
                            procedure draw_histogram_with_border(p_y_offset, p_fat: Integer);
                            procedure draw_vertical(p_x, p_y_offset, p_fat, p_pen_color: Integer);

                            Destructor Destroy; Override;
                          end; // c_histogram_255

  implementation
    uses Math, SysUtils
        , u_c_display;

    // -- c_histogram_255

    Constructor c_histogram_255.create_histogram_255(p_name: String;
        p_c_canvas: tCanvas; p_pen_color: Integer);
      begin
        Inherited create_basic_object(p_name);

        m_c_canvas:= p_c_canvas;
        m_pen_color:= p_pen_color;
        m_x_1:= k_no_x;
        m_x_2:= k_no_x;
      end; // create_histogram_255

    procedure c_histogram_255.reset_histogram;
      begin
        FillChar(m_histogram, SizeOf(m_histogram), 0);
      end; // reset_histogram

    procedure c_histogram_255.display_histogram;
      var l_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_total_count));
        for l_index:= 0 to 255 do
          if m_histogram[l_index]<> 0
            then display(Format('%3d  %7d ', [l_index, m_histogram[l_index]]));
      end; // display_histogram

    function c_histogram_255.f_total_count: Integer;
      var l_index: Integer;
      begin
        Result:= 0;
        for l_index:= 0 to 255 do
          Inc(Result, m_histogram[l_index]);
      end; // f_total_count

    procedure c_histogram_255.compute_max;
      var l_index: Integer;
      begin
        m_max:= m_histogram[0];

        for l_index:= 1 to 255 do
          m_max:= Max(m_max, m_histogram[l_index]);
      end; // compute_max

    procedure c_histogram_255.rescale_histogram(p_y_max: Integer);
      var l_index, l_max: Integer;
      begin
        if m_max<> 0
          then begin
              l_max:= 0;
              for l_index:= 0 to 255 do
              begin
                m_histogram[l_index]:= (m_histogram[l_index]* p_y_max) div m_max;
                l_max:= Max(l_max, m_histogram[l_index]);
              end;
              m_max:= l_max;
            end;
      end; // rescale_histogram

    procedure c_histogram_255.draw_histogram(p_y_min, p_y_max: Integer);

      function f_y(p_y: Integer): Integer;
        begin
          Result:= p_y_max- p_y;
        end; // f_y

      var l_index: Integer;
          l_save_pen_color: Integer;
          // l_save_brush

      begin // draw_histogram
        // display(Format('%-6s %4d %4d ', [m_name, p_y_min, p_y_max]));
        with m_c_canvas do
        begin
          l_save_pen_color:= Pen.Color;
          Pen.Color:= m_pen_color;

          // -- the base axis
          // MoveTo(0, p_y_max- p_y_min);
          //  LineTo(255, p_y_max- p_y_min);
          Rectangle(0, f_y(0), 257, f_y(m_max+ 2));

          MoveTo(0, f_y(p_y_min+ m_histogram[0]));
          for l_index:= 1 to 255 do
            LineTo(l_index, f_y(p_y_min+ m_histogram[l_index]));

          Pen.Color:= l_save_pen_color;
        end; // draw_histogram
      end; // draw_histogram

    procedure c_histogram_255.draw_histogram_with_border(p_y_offset, p_fat: Integer);

      function f_y(p_y: Integer): Integer;
        begin
          Result:= p_y_offset- p_y;
        end; // f_y

      procedure display_x1_x2;
        var l_save_brush_color: Integer;
        begin
          with m_c_canvas do
          begin
            l_save_brush_color:= Brush.Color;
            Brush.Color:= clWhite;
            if (m_x_1<> k_no_x) and (m_x_2<> k_no_x)
              then begin
                  // -- display 0..x_1
                  Rectangle(p_fat, f_y(m_max+ 2* p_fat- 1), m_x_1+ p_fat+ 1, f_y(0));
                  Rectangle(m_x_2+ p_fat, f_y(m_max+ 2* p_fat- 1), 255+ 2* p_fat- 1, f_y(0));
                end
              else
                if m_x_1<> k_no_x
                  then begin
                      // -- display 0..x_1
                      Rectangle(p_fat, f_y(m_max+ 2* p_fat- 1), m_x_1+ p_fat+ 1, f_y(0));
                    end
                  else
                    Rectangle(m_x_2+ p_fat, f_y(m_max+ 2* p_fat- 1), 255+ 2* p_fat- 1, f_y(0));
            Brush.Color:= l_save_brush_color;
          end; // with m_c_canvas
        end; // display_x1_x2

      var l_index: Integer;
          l_save_pen_color: Integer;

      begin // draw_histogram_with_border
        // display(Format('> draw_border %-6s y_offs=%4d max=%4d fat=%4d ',
        //    [m_name, p_y_offset, m_max, p_fat]));

        with m_c_canvas do
        begin
          l_save_pen_color:= Pen.Color;
          Pen.Color:= m_pen_color;

          // -- the frame
          // display(Format('f_y(m_max+ 2* p_fat- 1) %6d ', [f_y(m_max+ 2* p_fat- 1)]));

          Rectangle(0, f_y(m_max+ 2* p_fat- 1), 255+ 2* p_fat- 1, f_y(0));
          if (m_x_1<> k_no_x) or (m_x_2<> k_no_x)
            then display_x1_x2;

          // display(Format('zero x=fat=%5d   hist[0]=%4d  y=%5d', [p_fat, m_histogram[0], f_y(p_fat+ m_histogram[0])]));
          MoveTo(p_fat, f_y(p_fat+ m_histogram[0]));
          for l_index:= 1 to 255 do
            LineTo(p_fat+ l_index, f_y(p_fat+ m_histogram[l_index]));

          Pen.Color:= l_save_pen_color;
        end; // draw_histogram
        // display('< draw_border');
      end; // draw_histogram_with_border

    procedure c_histogram_255.draw_vertical(p_x, p_y_offset, p_fat, p_pen_color: Integer);

      function f_y(p_y: Integer): Integer;
        begin
          Result:= p_y_offset- p_y;
        end; // f_y

      var l_index: Integer;
          l_save_pen_mode: tPenMode;
          l_save_pen_color: Integer;

      begin // draw_vertical
        with m_c_canvas do
        begin
          l_save_pen_mode:= Pen.Mode;
          // Pen.Mode:= pmXor;
          l_save_pen_color:= Pen.Color;
          Pen.Color:= p_pen_color;

          MoveTo(p_x+ p_fat, f_y(m_max+ 2* p_fat- 1));
          LineTo(p_x+ p_fat, f_y(0));

          Pen.Color:= l_save_pen_color;
          Pen.Mode:= l_save_pen_mode;
        end; // draw_histogram
      end; // draw_vertical

    Destructor c_histogram_255.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_histogram_255
    end. // u_c_histogram_255
