// 002 u_graphic
// 15 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_graphic;
  interface
    uses Graphics;

    type t_rgb_color_type= (e_unknown_color, e_red_color, e_green_color, e_blue_color);

         // -- from Wintow
         t_RGB_triple= packed record
                         m_blue, m_green, m_red: Byte;
                        end;
         // -- this represents a line of 3 color bytes
         t_RGB_array= ARRAY[0..0] OF t_RGB_triple;
         // -- a pointer to this line
         t_pt_RGB_array= ^ t_RGB_array;

         t_RGB_virtual_triple= packed record
                                 m_virtual_blue, m_virtual_green, m_virtual_red: Integer;
                               end;
         t_RGB_virtual_array= ARRAY OF t_RGB_virtual_triple;

         t_float= Double;
         t_RGB_float_triple= record
                               m_float_blue, m_float_green, m_float_red: t_float;
                             end;
         t_RGB_float_array= ARRAY OF t_RGB_float_triple;

         t_hsv_triple= RECORD
                         m_color, m_saturation, m_intensity: Double;
                       END;
         t_hsv_float_array= ARRAY OF t_hsv_triple;              

    procedure copy_bitmap_to(p_c_source_bitmap, p_c_destination_bitmap: tBitmap);
    procedure or_bitmap_with(p_c_bitmap, p_c_bitmap_to_or: tBitmap);
    procedure saturate_non_white(p_c_bitmap: tBitmap; p_color: Integer);

    procedure bound_byte_range(var pv_byte_value: Integer);
    function f_bound_byte_range(var pv_byte_value: Integer): Boolean;

    function f_float_to_byte(p_float_value: t_float): Byte;
    function f_bound_float(p_float, p_min, p_max: t_float): t_float;
    function f_bound_byte(p_byte: Double; p_min, p_max: Byte): Byte;

    var g_gamma: array[0..255] of integer;
    procedure initialize_gamma;

    function f_rgb_to_hsv(CONST p_k_rgb_float_triple: t_RGB_float_triple): t_hsv_triple;
    FUNCTION f_hsv_to_rgb(CONST p_k_hsv_triple: t_hsv_triple): t_rgb_float_triple;
    function f_normalize_hsv(const pk_hsv_triple: t_hsv_triple): t_hsv_triple;

  implementation
    uses Math, SysUtils, u_math
       , u_c_display;

    procedure copy_bitmap_to(p_c_source_bitmap, p_c_destination_bitmap: tBitmap);
      begin
        with p_c_destination_bitmap do
        begin
          Width:= p_c_source_bitmap.Width;
          Height:= p_c_source_bitmap.Height;
          PixelFormat:= p_c_source_bitmap.PixelFormat;

          Canvas.Draw(0, 0, p_c_source_bitmap);
        end;
      end; // copy_bitmap

    procedure or_bitmap_with(p_c_bitmap, p_c_bitmap_to_or: tBitmap);
      var l_line, l_column: Integer;
          l_pixel, l_pixel_to_or: Integer;
      begin
        with p_c_bitmap do
          for l_line:= 0 to Height- 1 do
          begin
            for l_column:= 0 to Width- 1 do
            begin
              l_pixel:= Canvas.Pixels[l_line, l_column];
              l_pixel_to_or:= p_c_bitmap_to_or.Canvas.Pixels[l_line, l_column];
              if (l_pixel= clWhite) and (l_pixel_to_or<> clWhite)
                then begin
                    Canvas.Pixels[l_line, l_column]:= l_pixel_to_or;
                  end;
            end;
          end; // with p_c_bitmap, for l_line
      end; // or_bitmap_with

    procedure saturate_non_white(p_c_bitmap: tBitmap; p_color: Integer);
      var l_line, l_column: Integer;
          l_pixel, l_pixel_to_or: Integer;
      begin
        with p_c_bitmap do
          for l_line:= 0 to Height- 1 do
            for l_column:= 0 to Width- 1 do
            begin
              l_pixel:= Canvas.Pixels[l_line, l_column];
              if l_pixel<> clWhite
                then Canvas.Pixels[l_line, l_column]:= p_color;
            end;
      end; // saturate_non_white

    procedure bound_byte_range(var pv_byte_value: Integer);
      begin
        if pv_byte_value> 255
          then pv_byte_value:= 255
          else
            if pv_byte_value< 0
              then pv_byte_value:= 0;
      end; // check_byte_value

    function f_bound_float(p_float, p_min, p_max: t_float): t_float;
      begin
        Result:= p_float;
        if Result< p_min
          then Result:= p_min;
        if Result> p_max
          then Result:= p_max;
      end; // f_bound_float

    function f_bound_byte_range(var pv_byte_value: Integer): Boolean;
        // -- return True if no overflow
      begin
        if pv_byte_value> 255
          then begin
              pv_byte_value:= 255;
              Result:= False
            end
          else
            if pv_byte_value< 0
              then begin
                  pv_byte_value:= 0;
                  Result:= False
                end
              else Result:= True;
      end; // check_byte_value

    function f_bound_byte(p_byte: Double; p_min, p_max: Byte): Byte;
      var l_integer: Integer;
      begin
        l_integer:= Round(p_byte);
        if l_integer< p_min
          then Result:= p_min
          else
            if l_integer> p_max
              then Result:= p_max
              else Result:= l_integer;
      end; // f_bound_byte

    function f_float_to_byte(p_float_value: t_float): Byte;
      var l_integer: Integer;
      begin
        l_integer:= Round(p_float_value);
        if l_integer< 0
          then Result:= 0
          else
            if l_integer> 255
              then Result:= 255
              else Result:= l_integer;
      end; // f_float_to_byte

    // -- gamma correction

    procedure initialize_gamma;
      var l_index: Integer;
          l_value: Single;
      begin
        For l_index:= 0 to 255 do
        begin
          l_value:= l_index;
          g_gamma[l_index]:= round(f_power(l_value/ 255, 0.33)* 128);
        end;
      end; // initialize_gamma

    function f_float_min(p_one, p_two: t_float): t_float;
      begin
        if p_one<= p_two
          then Result:= p_one
          else Result:= p_two;
      end; // f_float_min

    function f_float_max(p_one, p_two: t_float): t_float;
      begin
        if p_one>= p_two
          then Result:= p_one
          else Result:= p_two;
      end; // f_float_min

    function f_float_equal(p_one, p_two: t_float): Boolean;
      begin
        Result:= Abs(p_one- p_two)< 1e-10;
      end; // f_float_equal

    // -- hsv rgb

    function f_rgb_to_hsv(CONST p_k_rgb_float_triple: t_RGB_float_triple): t_hsv_triple;
        // -- hue= 0 to 360 (corresponding to 0..360 degrees around hexcone)
        // -- saturation= 0 (shade of gray) to 1 (pure color)
        // -- intensity= 0 (black) to max {white)
      VAR l_intensity_min: t_float;
          l_intensity_range: t_float;
      BEGIN
        WITH p_k_rgb_float_triple, Result DO
        BEGIN
          // -- intensity:= Max(red, green, blue)
          // -- should be around [-255 .. 510]
          m_intensity:= f_float_max(m_float_red, f_float_max(m_float_green, m_float_blue));

          if m_intensity= 0
            then begin
                // -- saturation is 0 if r, g and b are all 0
                m_saturation:= 0;

                // -- achromatic: When saturation = 0, color is undefined
                // -- but arbitrarily assigned the value 0
                m_color:= 0;
                // display('*** acro');
              end
            else begin
                l_intensity_min:= f_float_min(m_float_red, f_float_min(m_float_green, m_float_blue));
                l_intensity_range:= m_intensity- l_intensity_min;

                // -- get the saturation as a percentage  range / max
                m_saturation:= l_intensity_range / m_intensity;

                if m_saturation= 0
                  then begin
                      m_color:= 0;
                      // -- also in white spots
                      // display('*** sat= 0 ');
                    end
                  else begin
                      // -- chromatic

                      // -- find the color with the maximum value
                      // -- and compute the spread with the other two colors
                      if m_float_red= m_intensity
                        then // -- the red is the greatest value
                             // -- this value becomes the intensity
                             // -- degrees between yellow and magenta
                             m_color:= (m_float_green- m_float_blue)* 60.0 / l_intensity_range
                        else
                          IF m_float_green= m_intensity
                            THEN // -- degrees between cyan and yellow
                                 m_color:= 120.0+ (m_float_blue- m_float_red)* 60.0/ l_intensity_range
                            else
                              if m_float_blue= m_intensity
                                THEN // -- degrees between magenta and cyan
                                     m_color:= 240.0+ (m_float_red- m_float_green)* 60.0/ l_intensity_range
                                else display_bug_halt('no_rgb_max');
                    end;
              end;

          IF m_color< 0
            THEN begin
                m_color:= m_color+ 360.0;
                if m_color> 360
                  then ; // display('*** 360');
              end;

          // -- normalize itensity to 1
          m_intensity:= m_intensity / 255;
        END; // WITH p_k_rgb_float_triple, Result
      END; // f_rgb_to_hsv

    FUNCTION f_hsv_to_rgb(CONST p_k_hsv_triple: t_hsv_triple): t_rgb_float_triple;

      procedure set_rgb(p_red, p_green, p_blue: t_float);
        begin
          With Result do
          begin
            m_float_red:= p_red;
            m_float_green:= p_green;
            m_float_blue:= p_blue;
          end;
        end; // set_rgb

      VAR l_quadrant_index: Integer;
          l_quadrant_fraction: t_float;
          p, q, t: t_float;

      BEGIN // f_hsv_to_rgb
        WITH p_k_hsv_triple, Result DO
        BEGIN
          IF m_saturation= 0
            THEN begin
                // -- achromatic: shades of gray
                m_float_red:= m_intensity;
                m_float_green:= m_intensity;
                m_float_blue:= m_intensity;
              end
            ELSE
              BEGIN
                // -- chromatic color

                // -- decompose the color into 6 quadrants:
                // --   the m_color is transformed into (quadrant, offset)
                l_quadrant_index:= Trunc(m_color / 60.0);
                if (l_quadrant_index< 0) or (l_quadrant_index> 5)
                  then l_quadrant_index:= 0;

                // -- l_quadrant_fraction in [0 .. 1[
                l_quadrant_fraction:= m_color- l_quadrant_index* 60.0;

                l_quadrant_fraction:= l_quadrant_fraction / 60.0;
                p:= m_intensity* (1- m_saturation);
                q:= m_intensity* (1- m_saturation* l_quadrant_fraction);
                t:= m_intensity* (1- m_saturation* (1- l_quadrant_fraction));
                CASE l_quadrant_index OF
                  0: // -- the red is the maximum
                     set_rgb(m_intensity, t, p);
                  1: set_rgb(q, m_intensity, p);
                  2: set_rgb(p, m_intensity, t);
                  3: set_rgb(p, q, m_intensity);
                  4: set_rgb(t, p, m_intensity);
                  5: set_rgb(m_intensity, p, q);
                  ELSE
                    // -- should never happen;
                    set_rgb(0, 0, 0)
                END; // case l_quadrant_index
              END
        end; // WITH p_k_hsv_triple, Result DO
      END; //  f_hsv_to_rgb

    function f_normalize_hsv(const pk_hsv_triple: t_hsv_triple): t_hsv_triple;
      begin
        Result:= pk_hsv_triple;
        with Result do
        begin
          if m_color< 0
            then m_color:= m_color+ 360
            else
              if m_color> 360
                then m_color:= m_color- (m_color /360)* 360;
          if m_saturation< 0
            then m_saturation:= 0
            else
              if m_saturation> 1
                then m_saturation:= 1;
          if m_intensity< 0
            then m_intensity:= 0
            else
              if m_intensity> 1
                then m_intensity:= 1;
        end; // with result
      end; // f_normalize_hsv

    end.



